{{
Microchip PCF8593 Real Time Clock driver.
Author: Clive J. Wakeham

The PCF8593 is a low power restricted memory version of the PCF8583 Real Time Clock.
Pin 3 of the PCF8593 is a negative going RESET pin which resets the I2C interface of the chip without affecting the time/date etc.
(On the PCF8583 pin 3 is an I2C address selection pin.)

                        +--\/--+
                  OSCI -|      |- VDD
                  OSCO -| PCF  |- INT~
                ~RESET -| 8593 |- SCL
                   VSS -|      |- SDA
                        +------+

The PCF8593 only has limited memory compared to the PCF8583, if you don't use the alarm function you get 8 bytes of RAM,
otherwise no spare memory.

Average power used at 5V is 4 microamps (at 2V is 1 microamp) where as the PCF8583 uses 50 microamps.
The PCF8593 will still keep time at 1 volt to 6 volts.

If the PCF8593 starts up without having the time/date set then register 00 will have the value of CC hex. If it was already set up previously
and is running then the value of register 00 will be otherwise.

The PCF8593 has a fixed I2C address of A2 hex. The registers of the PCF8593 is identical to the first 16 registers of the PCF8583.

The PCF8593 will increment the next register address after accessing the previous address, which I utilize for burst reading and writing.
(ie the ReadRegisterBlock and WriteRegisterBlock methods)

As per specifications for the PCF8593, any writing to the time or date registers will need to stop the clock to stop possible corruption of 
the registers. So the methods here for writing to the time or date registers will call the Stop_Clock method. So the Start_Clock method needs
to be called to get the clock going again. I have not set up an automatic call to Start_Clock after writing to the time or date registers,
so that all the necessary registers can be written to and set up before a Start_Clock call is made. But the calling program/method needs to
call the Start_Clock method manually.

All write methods will return an ACK if the write methods are called with the parameters within correct values otherwise they will return
an NAK value without writing to the registers.


Version History.
1.0 15/03/2011 -- Initial Build.
1.1 17/03/2011 -- Burst read and write methods.
1.2 22/04/2011 -- Scratchpad memory read and write methods.
1.3 23/04/2011 -- Alarm and Timer Mode setting methods.
1.4 20/05/2011 -- Changed calling methods for Time/Date.
1.5 22/05/2011 -- Updated all documentation for methods.
1.6 26/05/2011 -- Routine for storing and updating Full Year date.
1.7 28/05/2011 -- Major re-write of Alarm setting methods.

##################################################################################################
#                                           Registers                                            #
# ===============================================================================================#
# Control_status                  = $00                    Alarm Control                   = $08 #
# 100th_second                    = $01                    Alarm 100th seconds             = $09 #
# Seconds                         = $02                    Alarm Seconds                   = $0A #
# Minutes                         = $03                    Alarm Minutes                   = $0B #
# Hours                           = $04                    Alarm Hours                     = $0C #
# Year/Date                       = $05                    Alarm Date                      = $0D #
# Weekday/Month                   = $06                    Alarm Month                     = $0E #  
# Timer                           = $07                    Alarm Timer                     = $0F # 
##################################################################################################

function_mode_bits = %00  32.768 khz crystal between the OSCI and the OSCO pins.
function_mode_bits = %01  50 hz signal input on the OSCI pin. Nothing connected on OSCO pin.
function_mode_bits = %10  event timer signal input on the OSCI pin. Nothing connected on OSCO pin.

**************************************************************************************************************************
* RESTRICTIONS & LIMITATIONS.                                                                                            *
* +++++++++++++++++++++++++++                                                                                            *
* AM/PM not implemented, returns Hours in 24 hour format only. Easy to implement from the calling program/method.        *
* If the Alarm function of the PCF8593 is used then Year will only return the two insignificate bits for the year.       *
* Event Timer methods not implemented in this version of the PCF8593 object.                                             *
**************************************************************************************************************************
}}

Con
        device_address          = %10100010               'A2 Write -- A3 Read -- Fixed address
        SCL                     = 28                      'Standard pin for SCL, SDA = SCL + 1
        NAK                     = 0
        ACK                     = 1
                             
Obj
        I2C                     : "Basic_I2C_Driver"      

var
        byte                    value
        byte                    index                   
        byte                    value2[4]
        long                    RTC_value
        long                    Alarm_Status
        word                    Year

'Set up of the PCF8593 RTC Chip.
'================================================================================================       
pub RTC_Present : Ok
{{ Checks to see if RTC is connected and powered on. Returns a long as True or False. }}
RTC_value := I2C.devicePresent(SCL, device_address)                          'True or false read
return RTC_value                                                             'Return

pub RTC_Running : Ok
{{ Checks to see if the RTC is set up and programed. Easy way to check after power outage or brown-out
   RTC comes up with the value of CC hex in register 00 when powered on and was not running with reduced power.
   Returns a long as True or False. }}
value := I2C.ReadLocation(SCL, device_address, $00)                          'Check the register 00
if value <> $CC                                                              'Value of register not equal to CC hex?
 return TRUE                                                                 'Then RTC set and running.
else
 return FALSE                                                                'Else RTC not set.

pub Init_Crystal
{{ Sets up the Control Register to use a 32.768 KHz crystal on the OSCI and OSCO pins for timing,
   just changes the 4th and 5th bits of the register. }}
value := I2c.ReadLocation(SCL, device_address, $00)                          'Get old data from register 00
value &= %11001111                                                           'function mode %00 bits on bits 4 and 5
I2C.WriteLocation(SCL, device_address, $00, value)                           'Write modified data to register 00
return                                                                       'Return

pub Init_50hz_Signal
{{ Sets up the Control Register to use a 50Hz signal on the OSCI pin for timing. Just changes the 4th and 5th bits of the register. }}
value := I2c.ReadLocation(SCL, device_address, $00)                          'Get old data from Register 00 
value &= %11001111                                                           'clear bits 4 and 5
value |= %00010000                                                           'function mode %01 bits on bits 4 and 5 
I2C.WriteLocation(SCL, device_address, $00, value)                           'Write modified data to register 00
return 
                                                          
pub Start_Clock
{{ Puts a binary 0 in MSB of register 0 -- starts counting or restarts counting. Resets the alarm/timer bits also.
   This needs to be called if Set_Time or Set_Date is called since calling those methods will stop the clock. }}
value := I2c.ReadLocation(SCL, device_address, $00)                          'Get old data from Register 00 
value &= %00110000                                                           'Bit 7 = 0 to start clock
I2C.WriteLocation(SCL, device_address, $00, value)                           'Write modified data to Register 00 
return 

pub Stop_Clock
{{ Puts a binary 1 in MSB of register 0 -- stops the chip counting. }}
value := I2c.ReadLocation(SCL, device_address, $00)                          'Get old data from Register 00 
value |= %10000000                                                           'Bit 7 = 1 to stop clock
I2C.WriteLocation(SCL, device_address, $00, value)                           'Write modified data to Register 00 
return 

pub Read_Flags : Flag_Value
{{ Reads the Alarm and Timer flags and returns only those two bits. Returns a byte. }}
value := I2C.ReadLocation(SCL, device_address, $00)                          'Get data from Register 00 
value &= %00000011                                                           'First two bits.
return value

pub Reset_Flags
{{ Resets Alarm and Timer flags }}
value := I2C.ReadLocation(SCL, device_address, $00)                          'Get data from Register 00 
value &= %11111100                                                           'Clear bit 0 and bit 1
I2C.WriteLocation(SCL, device_address, $00, value)                           'Re-write the register 00
return 

'Set time and Set date methods.
'========================================================================================================================
pub Set_Time(Hour, Minute, Second)                                           'Set Hours, Minutes and Seconds
{{ Sets Hour, Minute and Seconds into respective registers. Hour must be in 24 hour format. AM/PM not used in this object.
   Returns a NAK if outside time ranges, or an ACK when done. }}
Stop_Clock                                                                   'As per specifications
if Hour >23 or Minute >59 or Second >59                                      'Simple check for hours under 24, minutes/seconds less than 60
 return NAK                                                                  'Return error 0
value2[0] := $00                                                             'Reset thousandth of second to zero
value2[1] := NumbertoBCD(Second)                                             'Set seconds
value2[2] := NumbertoBCD(Minute)                                             'Set minutes
value2[3] := NumbertoBCD(Hour)                                               'Set hours
WriteRegisterBlock(SCL, device_address, $01, 3)                              'Write hours, minutes, seconds, and reset thousandths of second. Register 01 starting register.
return ACK                                                                   'Return

pub Set_Date(Years, Month, Day, Weekday)                                      'Set Year, Month, Day and Weekday
{{ Sets the Year, Month, Day and Weekday. Year to be Word sized but only the two insignificant bits are used,
   all other values are byte sized. The PCF8593 only has a 2 bit year counter, with the value of 00 equal to a leap year.
   Checks to ensure that correct number of days equal to month, and 29 days of Feburary on falls on a leap year.
   Weekday are 0 for Monday through to 6 for Sunday. Returns a NAK if error, an ACK if okay. }}
Stop_Clock                                                                   'As per specifications
Year := (Years & %0000000000000011)                                          'RTC only does 4 yearly, so get rid of rest
Set_FullYear(Years, Year)                                                    'Store if possible full year in top byte locations
if Month == 2 and Day > 29                                                   'Only 28 or 29 days in February.
 return NAK
if Month == 2 and Day == 29 and Year <> 0                                    'Check for Leap year and Feburary with 29 days
 return NAK                                                                  'return NAK as error
if (Month == 9 or Month == 4 or Month == 6 or Month == 11) and Day > 30      'Check for 30 day months
 return NAK
if (Month == 1 or Month == 3 or Month == 5 or Month == 7) and Day > 31       'Check for 31 day months
 return NAK
if (Month == 8 or Month == 10 or Month == 12) and Day > 31                   'Check for 31 day months continued.
 return NAK
if Weekday > 6                                                               'Check for 7th day or less
 return NAK
Year ->= 2                                                                   'The MSB's are the year
value := Year.byte[1]                                                        'Highest byte for the year
Day := NumbertoBCD(Day)                                                      'Convert Day to BCD nibbles
Day |= value                                                                 'Year 2 highest bits and day lowest 6 bits
I2C.WriteLocation(SCL, device_address, $05, Day)                             'Write Year and Days into register 05
Month := NumbertoBCD(Month)                                                  'Convert Month to BCD
Month &= %00011111                                                           'Lowest 5 bits are the month
Weekday &= %00000111                                                         'Three bits to a weekday
Weekday ->= 3                                                                'Rotate to highest postion
value := Weekday | Month                                                     'Highest 3 bits are the weekend, lowest 5 bits month
I2C.WriteLocation(SCL, device_address, $06, value)                           'Write Month and Weekday into register 06
return ACK 

'Set Alarm modes
'========================================================================================================================
pub Activate_Alarm(Alarm_Control)
{{ If true then the alarm functions work, if false then registers $08 to $0F can be used as memory }}
value := I2c.ReadLocation(SCL, device_address, $00)                          'Get old data from Register 00 
Alarm_Status := Alarm_Control                                                'Set variable for later use.
if Alarm_Status                                                              'Turn on alarm?
 value |= %00000100                                                          'Bit 2 = 1 to activate alarm
else                                                                         'Turn off alarm?
 value &= %11111011                                                          'Bit 2 = 0 to deactivate alarm
I2C.WriteLocation(SCL, device_address, $00, value)                           'Write modified data to Register 00 
return 

pub Set_Daily_Alarm(Days, Hours, Minutes, Seconds, Interrupt)                'Mode 1 alarm function.
{{ Sets an alarm that occurs every day at a certain time. Interrupt is either 0 for FALSE or else is TRUE. All byte sized arguments}}
value := I2C.ReadLocation(SCL, device_address, $08)                          'Get old data from register 08
value &= %11001111                                                           'Lets clear bits 4 and 5 -- default mode 0
value |= %00010000                                                           'And set bit 4
if Interrupt == NAK                                                          'Is interrupt off?
 value &= %01111111                                                          'Clear bit 7
else                                                                         'Else interrupt on?
 value |= %10000000                                                          'Set bit 7
I2C.WriteLocation(SCL, device_address, $08, value)                           'Write modified data to Register 08
return ACK

pub Set_Weekday_Alarm(WeekDays, Hours, Minutes, Seconds, Interrupt)          'Mode 2 alarm function.
{{ Sets an alarm that will occur on certain days. Interrupt is either 0 for FALSE or else is TRUE. All byte sized arguments.
    the Weekday byte is structured with bit0 being Monday, bit1 being Tuesday etc etc}}
value := I2C.ReadLocation(SCL, device_address, $08)                          'Get old data from register 08
value &= %11001111                                                           'Lets clear bits 4 and 5 -- default mode 0
value |= %00100000                                                           'And set bit 5
if Interrupt == NAK                                                          'Is interrupt off?
 value &= %01111111                                                          'Clear bit 7
else                                                                         'Else interrupt on?
 value |= %10000000                                                          'Set bit 7
I2C.WriteLocation(SCL, device_address, $08, value)                           'Write modified data to register 08
I2C.WriteLocation(SCL, device_address, $0E, Weekdays)                        'Write Weekday data to register 0E
value := NumberToBCD(Hours)                                                  'Convert Hours to BCD
I2C.WriteLocation(SCL, device_address, $0C, value)                           'Write Hours data to register 0C
value := NumberToBCD(Minutes)                                                'Convert Minutes to BCD
I2C.WriteLocation(SCL, device_address, $0B, value)                           'Write Minutes data to register 0B
value := NumberToBCD(Seconds)                                                'Convert Seconds to BCD
I2C.WriteLocation(SCL, device_address, $0A, value)                           'Write Seconds data to register 0A  
return ACK

pub Set_Dated_Alarm(Months, Days, Hours, Minutes, Seconds, Interrupt)        'Mode 3 alarm function.
{{ Set specific month, day, hour, minutes and seconds for alarm. No year used.
   Interrupt is either 0 for FALSE or else is TRUE. All byte sized arguments
}}
value := I2C.ReadLocation(SCL, device_address, $08)                          'Get old data from register 08
value &= %11001111                                                           'Lets clear bits 4 and 5 -- default mode 0
value |= %00110000                                                           'And set bits 4 and 5
if Interrupt == NAK                                                          'Is interrupt off?
 value &= %01111111                                                          'Clear bit 7
else                                                                         'Else interrupt on?
 value |= %10000000                                                          'Set bit 7
I2C.WriteLocation(SCL, device_address, $08, value)                           'Write modified data to register 08
return ACK

'Set Timer mode
'=======================================================================================================================
pub Set_Timer_Mode(Mode, Interrupt, Alarm)
{{ Sets the Timer mode. Mode 0 is no timer. 1 is 100ths of a second, 2 is seconds, 3 is minutes, 4 is hours, 5 is days.
   Interrupt is TRUE for interrupt to activate, and FALSE for interrupt to be off.
   Alarm is TRUE for Timer Alarm to be enabled and FALSE for it to be disabled. }}
if mode > 5                                                                  'Only modes 0 to 5 allowed.
 return NAK                                                                  'Else return a NAK
value := I2C.ReadLocation(SCL, device_address, $08)                          'Get old data from register 08
value &= %11111000                                                           'Clear bits 0, 1 and 2 -- default mode 0
case Mode                                                                    'Greater than mode 0?
 1 : value |= %00000001                                                      'Mode 1?
 2 : value |= %00000010                                                      'Mode 2?
 3 : value |= %00000011                                                      'Mode 3?
 4 : value |= %00000100                                                      'Mode 4?
 5 : value |= %00000101                                                      'Mode 5?
if Interrupt                                                                 'Is interrupt on?
 value |= %00001000                                                          'Set bit 3
else                                                                         'Or off?
 value &= %11110111                                                          'Clear bit 3
if Alarm                                                                     'Alarm on?
 value |= %01000000                                                          'Set bit 6
else                                                                         'Or off?
 value &= %10111111                                                          'Clear bit 6
I2C.WriteLocation(SCL, device_address, $08, value)                           'Write modified data to register 08
return ACK

'Read Time and Date methods
'======================================================================================================================
pub Read_Time : time
{{ Read the current time and return decimal number values as a long. Any read to addresses 0 to 7 will put the current values of those
   registers into HOLD latches to stop errors occurring during carry-over if reading from those registers. Hours will be in 24 hour format.
   Returns a long with the highest byte hours, then minutes, then seconds, and then hundredths of seconds. }}
ReadRegisterBlock(SCL, device_address, $01, 3)                               '4 consecutive register location read from register 01
RTC_value.byte[0] := BCDtoNumber(value2[0])                                  'Convert value from chip to decimals for 100ths of second
RTC_value.byte[1] := BCDtoNumber(value2[1])                                  'Convert value from chip to decimals for seconds
RTC_value.byte[2] := BCDtoNumber(value2[2])                                  'Convert value from chip to decimals for minutes
RTC_value.byte[3] := BCDtoNumber(value2[3])                                  'Convert value from chip to decimals for hours
return RTC_value                                                             'Return long value

pub Read_Date : shortdate
{{ Read the current date and return decimal number values as a long. Any read to addresses 0 to 7 will put the current values of those
   registers into HOLD latches to stop errors occurring during carry-over if reading from those registers. The year value will be only 2 bits,
   ie 00 , 01, 10, or 11. The calling program will need to expand this into the correct year. Or use Read_FullYear if alarm not used.
   Returns a long with the highest byte as Weekday, then Day, Month and then Year. }}
ReadRegisterBlock(SCL, device_address, $05, 1)                               '2 consecutive register location read from register 05
value2[2] := value2[0]                                                       'Save the data over to keep the Day data     
value2[0] >>= 6                                                              'Lets clear the day off and put the year in the first 2 bits
value2[2] &= %00111111                                                       'Lets clear the year off
value2[2] := BCDtoNumber(value2[2])                                          'Convert BCD value from chip to decimal for Day
value2[3] := value2[1]                                                       'Save the value over to keep the weekday value    
value2[3] >>= 5                                                              'Clear off the Month and move the weekday into the first 3 bits
value2[1] &= %00011111                                                       'Clear off the weekday value
value2[1] := BCDtoNumber(value2[1])                                          'Convert BCD value from chip to decimal for Month
RTC_value.byte[0] := value2[0]                                               'Store Year -- the two insignificate bits only
RTC_value.byte[1] := value2[1]                                               'Store Month 
RTC_value.byte[2] := value2[2]                                               'Store Day 
RTC_value.byte[3] := value2[3]                                               'Store Weekday
Update_FullYear                                                              'Update the Full Year date if necessary etc
return RTC_value                                                             'Return long value

PUB Read_FullYear : Years
{{ Checks to see if the highest register addresses are not used for alarms, and then gets the complete year data.
   Returns a word with 0 for error, or word with 12 bit resolution of year. }}
if Alarm_Status                                                              'Check if the alarm is on
 return NAK                                                                  'If so, return error.
ReadRegisterBlock(SCL, device_address, $0E, 1)                               'Read last register addresses.
Year.byte[0] := value2[0]                                                    'Retrieve low byte
Year.byte[1] := (value2[1] & %00001111)                                      'Retrieve high byte, mask out top four bits
return Year

PUB Update_FullYear
{{ Compares the short year date from the PCF8593 to the value stored with FullYear date stored in locations 0E and 0F of the chip.
   If the figures don't match then the FullYear figure is incremented until the figures match. -- This is not a fail-safe way of keeping
   the FullYear date correct. Returns an 0 for error, or 1 for okay. }}
if Alarm_Status                                                              'Alarm turned on?
 return NAK                                                                  'If so, return error.
value := I2C.ReadLocation(SCL, device_address, $05)                          'Get year data from RTC chip. Register $05
value >>= 6                                                                  'Lets clear the day off and put the year in the first 2 bits
ReadRegisterBlock(SCL, device_address, $0E, 1)                               'Read last register addresses.
Year.byte[0] := value2[0]                                                    'Retrieve low byte
Year.byte[1] := (value2[1] & %00001111)                                      'Retrieve high byte, mask out top four bits
value2[1] >>= 6                                                              'Move two highest bits to lowest position
repeat until value == value2[1]                                              'Correction loop
 Year++                                                                      'Increment the Full Year
 value := ((value + 1) & %00000011)                                          'Increment the value variable
Set_FullYear(Year, value2[1])                                                'Put new figure in memory
return ACK                                                                   'Return all good.

'Direct access to memory on chip
'========================================================================================================================
pub Read_Memory(Location) : memory_value
{{ Read any of the eight bytes of scratchpad RAM. Returns a byte value. }}
'Location #>= 0                                                               'Lowest location 0
'Location <#= 7                                                               'Highest location 7
value := I2C.ReadLocation(SCL, device_address, Location)               'So read to location which is from 8 to 13
return value                                                                 'Return byte value

pub Write_Memory(Location, memory_value)
{{ Writes to any of the six bytes of scratchpad RAM, only if the Alarm is off.
   If the Alarm is off, the set date function puts the full date to locations 0E and 0F. So no overwriting these locations.
   Returns a NAK if error, or an ACK for okay. }}
'if Alarm_Status                                                              'No writing of the memory if Alarm is on.
' return NAK                                                                  'Return error
'Location #>= 0                                                               'Lowest location 0
'Location <#= 5                                                               'Highest location 5
I2C.WriteLocation(SCL, device_address, Location, memory_value)         'So write to location which is from 8 to 13
return ACK

'Private methods.
'========================================================================================================================
PRI Set_FullYear(Years, Old_Year_Value)
{{ Checks the full year value and if the alarm functions are not being used then it stores it in register locations 0E and 0F.
   The format of the stored "word" of the full year date is the two highest bits are a copy of the year data from the PCF8593
   when this function was called or Update_FullYear was called last. And then two zeros and then the lower 12 bits of date.}}
if Alarm_Status                                                              'Check if the alarm is on
 return NAK                                                                  'If so, return.
Years &= %0000111111111111                                                   '12 bits resolution (Up to year 4095!!!)
value := Old_Year_Value.byte[0] << 6                                         'Move two lowest bits to highest position
value2[0] := Years.byte[0]                                                   'Store low byte
value2[1] := (Years.byte[1] | value)                                         'Store high byte
WriteRegisterBlock(SCL, device_address, $0E, 1)                              'Store the values in registers 0E and 0F
return

PRI NumberToBCD(number)
{{ Converts a binary number to a BCD number by Kyle Crane and Kwabena W. Agyeman }}
return (((number / 10) << 4) + (number // 10))        

PRI BCDToNumber(BCD)
{{ Converts a BCD number to a binary number by Kyle Crane and Kwabena W. Agyeman }}
return (((BCD >> 4) * 10) + (BCD & $F))

PRI ReadRegisterBlock(SCL_pin, device_address2, register_address, count)
{{ Reads multiple registers in sequence, advancement of James's Extra Bits of Michael Green's Basic I2C Driver}}
  I2C.start(SCL_pin)
  I2C.write(SCL_pin, device_address2 | 0)
  I2C.write(SCL_pin, register_address)
  I2C.start(SCL_pin)
  I2C.write(SCL_pin, device_address2 | 1)  
  repeat index from 0 to count
   value2[index] := I2C.read(SCL_pin, NAK)
  I2C.stop(SCL_pin)
  return ACK

PRI WriteRegisterBlock(SCL_pin, device_address2, register_address, count)
{{ Writes multiple registers in sequence, advancement of James's Extra Bits of Michael Green's Basic I2C Driver }}
  I2C.start(SCL_pin)
  I2C.write(SCL_pin, device_address2)
  I2C.write(SCL_pin, register_address)
  repeat index from 0 to count
   I2C.write(SCL_pin, value2[index])
  I2C.stop(SCL_pin)
  return ACK

{{

}}

{{
********************************************************************************************************************************
*                                                   TERMS OF USE: MIT License                                                  *                                                            
********************************************************************************************************************************
*Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    * 
*files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    *
*modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software*
*is furnished to do so, subject to the following conditions:                                                                   *
*                                                                                                                              *
*The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.*
*                                                                                                                              *
*THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          *
*WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         *
*COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   *
*ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         *
********************************************************************************************************************************
}}
    